/* global include enigma */ 
class WebsyQlikObjectManagerLight {
  constructor (options) {
    const defaults = {      
      helpEvent: 'mouseover',
      actions: []
    }
    this.app = null
    this.paused = false
    this.supportedChartTypes = []
    this.activeViews = []
    this.chartLibrary = {}
    this.globalObjectsLoaded = false    
    if (options.visualisationPlugins && options.visualisationPlugins.length > 0) {
      for (let i = 0; i < options.visualisationPlugins.length; i++) {
        this.registerVisualisation(options.visualisationPlugins[i].id, options.visualisationPlugins[i].definition)
      }
    }
    this.options = Object.assign({}, defaults, options)            
  }
  buildChildElement (elementId, text) {
    let html = `
      <article id='${elementId}_vis' class='websy-vis-article'></article>
      <div id='${elementId}_loading' class='websy-loading-container'><div class='websy-ripple'><div></div><div></div></div></div>
    `
    if (text && text !== '') {
      html += `
        <i class='websy-vis-help-listener' data-element='${elementId}'></i>
        <div id='${elementId}_help' class='websy-vis-help'><span>${text || ''}</span></div>        
      `
    }
    return html
  }  
  static mergeObjects () {    
    // Variables
    let extended = {}
    let deep = false
    let i = 0

    // Check if a deep merge
    if (typeof arguments[0] === 'boolean') {
      deep = arguments[0]
      i++
    }

    // Merge the object into the extended object
    let merge = function (obj) {
      for (let prop in obj) {
        if (obj.hasOwnProperty(prop)) {
          if (deep && Object.prototype.toString.call(obj[prop]) === '[object Object]') {
            // If we're doing a deep merge and the property is an object
            extended[prop] = this.mergeObjects(true, extended[prop], obj[prop])
          } 
          else {
            // Otherwise, do a regular merge
            if (Array.isArray(extended[prop]) && Array.isArray(obj[prop])) {
              extended[prop] = extended[prop].concat(obj[prop])
            }
            else {
              extended[prop] = obj[prop]  
            }            
          }
        }
      }
    }
    // Loop through each object and conduct a merge
    for (; i < arguments.length; i++) {
      merge(arguments[i])
    }
    return extended
  }
  init () {
    return new Promise((resolve, reject) => {
      this.prep()
      this.connectToApp().then(resolve, reject)
    })
  }
  pause () {
    this.paused = true
  }
  play (resume) {
    this.paused = false
    if (resume === true) {
      this.loadObjects('global')
      for (let i = 0; i < this.activeViews.length; i++) {
        this.loadObjects(this.activeViews[i])
      }
    }        
  }
  prep () {
    for (let view in this.options.views) {
      // sort out the elements in each view
      for (let o = 0; o < this.options.views[view].objects.length; o++) {
        let config = this.options.views[view].objects[o]
        let el = document.getElementById(config.elementId)
        if (el) {
          el.innerHTML += this.buildChildElement(config.elementId, config.help)
          if (config.help && config.help !== '') {
            el.addEventListener(this.options.helpEvent, this.handleEvent.bind(this))
            el.addEventListener('mouseout', this.handleEvent.bind(this))
          }
        }
      }
    }
    // setup  the event listeners for the actions
    // actions should not have a visualisation in the same property structure
    for (let a = 0; a < this.options.actions.length; a++) {      
      let el = document.getElementById(this.options.actions[a].elementId)
      if (el) {
        el.addEventListener(this.options.actions[a].event, () => {                                
          for (let i = 0; i < this.options.actions[a].items.length; i++) {
            let item = this.options.actions[a].items[i]
            if (item.field) {
              this.app.getField(item.field).then(field => {
                field[item.method](...item.params)
              })
            }
            else {
              this.app[item.method](...item.params)
            }
          }          
        })
      }
    }
  }
  connectToApp () {
    return new Promise((resolve, reject) => {
      // check for enigma.js      
      const originalId = this.options.enigmaConfig.app
      this.options.enigmaConfig.app = this.normalizeId(this.options.enigmaConfig.app)
      if (typeof enigma === 'undefined') {
        reject({
          error: 'Enigma.js not found.'
        })
        return
      }
      if (typeof this.options.enigmaSchema === 'undefined') {
        reject({
          error: 'enigmaSchema property not found.'
        })
        return
      }
      let config = {
        schema: this.options.enigmaSchema,
        url: this.options.enigmaConfig.url
      }
      let session = enigma.create(config)
      this.session = session
      session.open().then(global => {
        global.getActiveDoc().then(app => {          
          if (app) {
            return app
          }
          else {
            return global.openDoc(originalId)
          }
        }, err => {
          console.log(err)
          return global.openDoc(originalId)
        }).then(app => {
          this.app = app
          if (this.options.views.global) {
            this.executeActions('global').then(() => {
              // this.loadObjects('global')
              resolve()
            })            
          }
          else {
            resolve()  
          }  
        }, err => {
          reject(err)
        })
      }, err => {
        reject(err)
      })
      session.on('traffic:received', (data) => {
        if (typeof data.suspend !== 'undefined') {
          this.sessionSuspended()
        }        
      })
      session.on('notification:*', (eventName, data) => {        
        if (eventName === 'OnAuthenticationInformation') {
          if (data.mustAuthenticate === true && this.options.enigmaConfig.authUrl) {
            window.location = this.options.enigmaConfig.authUrl
          }
          else if (data.mustAuthenticate === false) {
            this.user = {
              userDirectory: data.userDirectory,
              userId: data.userId
            }
          }
        }
        else {
          this.sessionOnNotification(data)
        }
      })
      session.on('suspended', this.sessionSuspended.bind(this))
      session.on('resumed', this.sessionResumed.bind(this))
    })        
  }
  loadView (view) {
    if (view === '') {
      return
    }
    if (this.activeViews.indexOf(view) === -1 && view !== 'global') {
      this.activeViews.push(view)
    }    
    this.executeActions(view).then(() => {    
      if (this.globalObjectsLoaded === false) {
        this.loadObjects('global')
        this.globalObjectsLoaded = true
      }
      this.loadObjects(view)
    })
  }
  // executeActions(view){
  //   return new Promise((resolve, reject) => {
  //     if (!this.options.views[view].actions) {
  //       resolve()
  //     }
  //     let actionList = []
  //     for (var i = 0; i < this.options.views[view].actions.length; i++) {
  //       let item = this.options.views[view].actions[i]
  //       if (item.field) {
  //         actionList.push(this.app.getField(item.field).then(field=>{
  //           field[item.method](...item.params)
  //         }))
  //       }
  //       else {
  //         actionList.push(this.app[item.method](...item.params))
  //       }        
  //     }
  //     Promise.all(actionList).then(resolve, reject)
  //   })
  // }
  executeAction (index, actionList, callbackFn) {
    let item = actionList[index]
    if (item.field) {
      this.app.getField(item.field).then(field => {
        field[item.method](...item.params).then(() => {
          if (item.lock === true) {
            field.lock().then(() => {
              index++
              if (index === actionList.length) {
                callbackFn()
              }
              else {
                this.executeAction(index, actionList, callbackFn)
              }
            })
          }
          else {
            index++
            if (index === actionList.length) {
              callbackFn()
            }
            else {
              this.executeAction(index, actionList, callbackFn)
            }
          }          
        })
      })
    }
    else {
      this.app[item.method](...item.params).then(() => {
        index++
        if (index === actionList.length) {
          callbackFn()
        }
        else {
          this.executeAction(index, actionList, callbackFn)
        }
      })
    }
  }
  executeActions (view) {    
    return new Promise((resolve, reject) => {
      if (!this.options.views[view].actions || this.options.views[view].actions.length === 0) {
        resolve()
      }
      this.executeAction(0, this.options.views[view].actions, resolve)
    })
  }
  loadObjects (view) {    
    if (this.paused === true) {
      return
    }    
    let objList = this.options.views[view].objects
    if (objList && objList.length > 0) {
      for (let i = 0; i < objList.length; i++) {                
        if (objList[i].objectId) {
          objList[i].attached = true
          if (objList[i].vis && objList[i].vis.render) {
            objList[i].vis.render()  
          }
          else if (objList[i].vis.render) {
            objList[i].render()
          }          
        }
        else if (objList[i].definition) {
          this.createObjectFromDefinition(objList[i])
        }        
      }
    }
  }
  closeObjects (view) {
    if (view === '') {
      return
    }
    const viewIsActive = this.activeViews.indexOf(view)
    if (viewIsActive !== -1) {
      this.activeViews.splice(viewIsActive, 1)
    }
    let objList = this.options.views[view].objects
    if (objList && objList.length > 0) {
      for (let i = 0; i < objList.length; i++) {
        if (objList[i].vis) {
          objList[i].attached = false
          if (objList[i].vis.close) {
            objList[i].vis.close()  
          }
        }
        else if (objList[i].objectId) {
          if (objList[i].close) {
            objList[i].close()  
          }
          this.destroyObjectFromId(objList[i])          
        }
      }
    }
  }  
  handleEvent (event) {
    if (event.target.classList.contains('websy-vis-help-listener')) {
      let elementId = event.target.attributes['data-element']
      if (elementId.value) {
        this.toggleHelp(`${elementId.value}_help`)
      }
    }
  }
  createObjectFromDefinition (objectConfig) {
    /* global objectConfig */ 
this.app.createSessionObject(objectConfig.definition).then(model => {
  objectConfig.objectId = model.id
  objectConfig.attached = true
  if (this.supportedChartTypes.indexOf(objectConfig.definition.qInfo.qType) !== -1) {
    objectConfig.vis = new this.chartLibrary[objectConfig.definition.qInfo.qType](`${objectConfig.elementId}_vis`, {
      model, 
      def: objectConfig.definition,
      app: this.app
    })
    model.on('changed', () => {
      if (objectConfig.attached === true && this.paused === false) {
        objectConfig.vis.render()
      }
    })
  }
  else if (objectConfig.render && typeof objectConfig.render === 'function') {
    objectConfig.vis = {}
    objectConfig.render(objectConfig, model)
    model.on('changed', () => {
      if (objectConfig.attached === true && this.paused === false) {
        objectConfig.render(objectConfig, model)
      }
    })
  }
})

  }  
  destroyObjectFromId (objectConfig) {
    let hostEl = document.getElementById(`${objectConfig.elementId}_vis`)
    if (hostEl) {
      hostEl.innerHTML = ''
    }
    this.app.destroySessionObject(objectConfig.objectId)
  }
  detachObject (objectConfig) {
    objectConfig.attached = false
  }
  normalizeId (id) {
    return id.replace(/\s:\\\//, '-')
  }
  sessionOnNotification (event) {
    console.log('Notification')
    console.log(event)
  }
  sessionOnTraffic (event) {
    console.log('Traffic')
    console.log(event)
  }
  sessionResumed (event) {    
    console.log('Session Resumed')
    console.log(event)
  }
  sessionSuspended (event) {
    console.log('Session suspended')
    console.log(event)
  }
  showHelp (elementId) {
    let el = document.getElementById(elementId)
    if (el) {
      el.classList.add('active')
    }
  }
  hideHelp (elementId) {
    let el = document.getElementById(elementId)
    if (el) {
      el.classList.remove('active')
    }
  }
  toggleHelp (elementId) {
    let el = document.getElementById(elementId)
    if (el) {
      el.classList.toggle('active')
    }
  }
  onError (err) {
    console.log(err)
  }
  onClose (msg) {}
  resize () {
    for (let i = 0; i < this.activeViews.length; i++) {
      this.resizeObjects(this.activeViews[i])
    }
  }
  resizeObjects (view) {
    if (view === '') {
      return
    }    
    let objList = this.options.views[view].objects
    if (objList && objList.length > 0) {
      for (let i = 0; i < objList.length; i++) {                
        if (objList[i].objectId) {          
          if (objList[i].vis && objList[i].vis.resize) {
            objList[i].vis.resize()  
          }
          else if (objList[i].resize) {
            objList[i].resize()
          }          
        }                
      }
    }
  }
  registerVisualisation (name, classDef) {
    if (name.indexOf(/\s/) !== -1) {
      console.log('Failed to register Chart Extension. Chart name must not contain spaces.')
      return
    }
    if (this.supportedChartTypes.indexOf(name) !== -1) {
      console.log('Failed to register Chart Extension. Chart name already exists.')
      return
    }
    this.supportedChartTypes.push(name)
    this.chartLibrary[name] = classDef
  }
}
